"""
Affine transforms implemented on torch tensors, and
requiring only one interpolation
"""

import math
import random
import torch as th

from ..utils import th_affine2d, th_random_choice


class RandomAffine(object):

    def __init__(self, 
                 rotation_range=None, 
                 translation_range=None,
                 shear_range=None, 
                 zoom_range=None,
                 interp='bilinear',
                 lazy=False):
        """
        Perform an affine transforms with various sub-transforms, using
        only one interpolation and without having to instantiate each
        sub-transform individually.

        Arguments
        ---------
        rotation_range : one integer or float
            image will be rotated randomly between (-degrees, degrees) 

        translation_range : a float or a tuple/list with 2 floats between [0, 1)
            first value:
                image will be horizontally shifted between 
                (-height_range * height_dimension, height_range * height_dimension)
            second value:
                Image will be vertically shifted between 
                (-width_range * width_dimension, width_range * width_dimension)

        shear_range : float
            image will be sheared randomly between (-degrees, degrees)

        zoom_range : list/tuple with two floats between [0, infinity).
            first float should be less than the second
            lower and upper bounds on percent zoom. 
            Anything less than 1.0 will zoom in on the image, 
            anything greater than 1.0 will zoom out on the image.
            e.g. (0.7, 1.0) will only zoom in, 
                 (1.0, 1.4) will only zoom out,
                 (0.7, 1.4) will randomly zoom in or out

        interp : string in {'bilinear', 'nearest'} or list of strings
            type of interpolation to use. You can provide a different
            type of interpolation for each input, e.g. if you have two
            inputs then you can say `interp=['bilinear','nearest']

        """
        self.transforms = []
        if rotation_range is not None:
            rotation_tform = RandomRotate(rotation_range, lazy=True)
            self.transforms.append(rotation_tform)

        if translation_range is not None:
            translation_tform = RandomTranslate(translation_range, lazy=True)
            self.transforms.append(translation_tform)

        if shear_range is not None:
            shear_tform = RandomShear(shear_range, lazy=True)
            self.transforms.append(shear_tform) 

        if zoom_range is not None:
            zoom_tform = RandomZoom(zoom_range, lazy=True)
            self.transforms.append(zoom_tform)

        self.interp = interp
        self.lazy = lazy

        if len(self.transforms) == 0:
            raise Exception('Must give at least one transform parameter')

    def __call__(self, *inputs):
        # collect all of the lazily returned tform matrices
        tform_matrix = self.transforms[0](inputs[0])
        for tform in self.transforms[1:]:
            tform_matrix = tform_matrix.mm(tform(inputs[0])) 
        self.tform_matrix = tform_matrix

        if self.lazy:
            return tform_matrix
        else:
            outputs = Affine(tform_matrix,
                             interp=self.interp)(*inputs)
            return outputs


class Affine(object):

    def __init__(self, 
                 tform_matrix,
                 interp='bilinear'):
        """
        Perform an affine transforms with various sub-transforms, using
        only one interpolation and without having to instantiate each
        sub-transform individually.

        Arguments
        ---------
        tform_matrix : a 2x3 or 3x3 matrix
            affine transformation matrix to apply

        interp : string in {'bilinear', 'nearest'} or list of strings
            type of interpolation to use. You can provide a different
            type of interpolation for each input, e.g. if you have two
            inputs then you can say `interp=['bilinear','nearest']

        """
        self.tform_matrix = tform_matrix
        self.interp = interp

    def __call__(self, *inputs):
        if not isinstance(self.interp, (tuple,list)):
            interp = [self.interp]*len(inputs)
        else:
            interp = self.interp

        outputs = []
        for idx, _input in enumerate(inputs):
            input_tf = th_affine2d(_input,
                                   self.tform_matrix,
                                   mode=interp[idx])
            outputs.append(input_tf)
        return outputs if idx > 1 else outputs[0]


class AffineCompose(object):

    def __init__(self, 
                 transforms,
                 interp='bilinear'):
        """
        Apply a collection of explicit affine transforms to an input image,
        and to a target image if necessary

        Arguments
        ---------
        transforms : list or tuple
            each element in the list/tuple should be an affine transform.
            currently supported transforms:
                - Rotate()
                - Translate()
                - Shear()
                - Zoom()

        interp : string in {'bilinear', 'nearest'} or list of strings
            type of interpolation to use. You can provide a different
            type of interpolation for each input, e.g. if you have two
            inputs then you can say `interp=['bilinear','nearest']

        """
        self.transforms = transforms
        self.interp = interp
        # set transforms to lazy so they only return the tform matrix
        for t in self.transforms:
            t.lazy = True

    def __call__(self, *inputs):
        # collect all of the lazily returned tform matrices
        tform_matrix = self.transforms[0](inputs[0])
        for tform in self.transforms[1:]:
            tform_matrix = tform_matrix.mm(tform(inputs[0])) 

        if not isinstance(self.interp, (tuple,list)):
            interp = [self.interp]*len(inputs)
        else:
            interp = self.interp

        outputs = []
        for idx, _input in enumerate(inputs):
            input_tf = th_affine2d(_input,
                                   tform_matrix,
                                   mode=interp[idx])
            outputs.append(input_tf)
        return outputs if idx > 1 else outputs[0]


class RandomRotate(object):

    def __init__(self, 
                 rotation_range,
                 interp='bilinear',
                 lazy=False):
        """
        Randomly rotate an image between (-degrees, degrees). If the image
        has multiple channels, the same rotation will be applied to each channel.

        Arguments
        ---------
        rotation_range : integer or float
            image will be rotated between (-degrees, degrees) degrees

        interp : string in {'bilinear', 'nearest'} or list of strings
            type of interpolation to use. You can provide a different
            type of interpolation for each input, e.g. if you have two
            inputs then you can say `interp=['bilinear','nearest']

        lazy    : boolean
            if true, only create the affine transform matrix and return that
            if false, perform the transform on the tensor and return the tensor
        """
        self.rotation_range = rotation_range
        self.interp = interp
        self.lazy = lazy

    def __call__(self, *inputs):
        degree = random.uniform(-self.rotation_range, self.rotation_range)

        if self.lazy:
            return Rotate(degree, lazy=True)(inputs[0])
        else:
            outputs = Rotate(degree,
                             interp=self.interp)(*inputs)
            return outputs


class RandomChoiceRotate(object):

    def __init__(self, 
                 values,
                 p=None,
                 interp='bilinear',
                 lazy=False):
        """
        Randomly rotate an image from a list of values. If the image
        has multiple channels, the same rotation will be applied to each channel.

        Arguments
        ---------
        values : a list or tuple
            the values from which the rotation value will be sampled

        p : a list or tuple the same length as `values`
            the probabilities of sampling any given value. Must sum to 1.

        interp : string in {'bilinear', 'nearest'} or list of strings
            type of interpolation to use. You can provide a different
            type of interpolation for each input, e.g. if you have two
            inputs then you can say `interp=['bilinear','nearest']

        lazy    : boolean
            if true, only create the affine transform matrix and return that
            if false, perform the transform on the tensor and return the tensor
        """
        if isinstance(values, (list, tuple)):
            values = th.FloatTensor(values)
        self.values = values
        if p is None:
            p = th.ones(len(values)) / len(values)
        else:
            if abs(1.0-sum(p)) > 1e-3:
                raise ValueError('Probs must sum to 1')
        self.p = p
        self.interp = interp
        self.lazy = lazy

    def __call__(self, *inputs):
        degree = th_random_choice(self.values, p=self.p)

        if self.lazy:
            return Rotate(degree, lazy=True)(inputs[0])
        else:
            outputs = Rotate(degree,
                             interp=self.interp)(*inputs)
            return outputs


class Rotate(object):

    def __init__(self, 
                 value,
                 interp='bilinear',
                 lazy=False):
        """
        Randomly rotate an image between (-degrees, degrees). If the image
        has multiple channels, the same rotation will be applied to each channel.

        Arguments
        ---------
        rotation_range : integer or float
            image will be rotated between (-degrees, degrees) degrees

        interp : string in {'bilinear', 'nearest'} or list of strings
            type of interpolation to use. You can provide a different
            type of interpolation for each input, e.g. if you have two
            inputs then you can say `interp=['bilinear','nearest']

        lazy    : boolean
            if true, only create the affine transform matrix and return that
            if false, perform the transform on the tensor and return the tensor
        """
        self.value = value
        self.interp = interp
        self.lazy = lazy

    def __call__(self, *inputs):
        if not isinstance(self.interp, (tuple,list)):
            interp = [self.interp]*len(inputs)
        else:
            interp = self.interp

        theta = math.pi / 180 * self.value
        rotation_matrix = th.FloatTensor([[math.cos(theta), -math.sin(theta), 0],
                                          [math.sin(theta), math.cos(theta), 0],
                                          [0, 0, 1]])
        if self.lazy:
            return rotation_matrix
        else:
            outputs = []
            for idx, _input in enumerate(inputs):
                input_tf = th_affine2d(_input,
                                       rotation_matrix,
                                       mode=interp[idx],
                                       center=True)
                outputs.append(input_tf)
            return outputs if idx > 1 else outputs[0]


class RandomTranslate(object):

    def __init__(self, 
                 translation_range,
                 interp='bilinear',
                 lazy=False):
        """
        Randomly translate an image some fraction of total height and/or
        some fraction of total width. If the image has multiple channels,
        the same translation will be applied to each channel.

        Arguments
        ---------
        translation_range : two floats between [0, 1) 
            first value:
                fractional bounds of total height to shift image
                image will be horizontally shifted between 
                (-height_range * height_dimension, height_range * height_dimension)
            second value:
                fractional bounds of total width to shift image 
                Image will be vertically shifted between 
                (-width_range * width_dimension, width_range * width_dimension)

        interp : string in {'bilinear', 'nearest'} or list of strings
            type of interpolation to use. You can provide a different
            type of interpolation for each input, e.g. if you have two
            inputs then you can say `interp=['bilinear','nearest']

        lazy    : boolean
            if true, only create the affine transform matrix and return that
            if false, perform the transform on the tensor and return the tensor
        """
        if isinstance(translation_range, float):
            translation_range = (translation_range, translation_range)
        self.height_range = translation_range[0]
        self.width_range = translation_range[1]
        self.interp = interp
        self.lazy = lazy

    def __call__(self, *inputs):
        # height shift
        random_height = random.uniform(-self.height_range, self.height_range)
        # width shift
        random_width = random.uniform(-self.width_range, self.width_range)

        if self.lazy:
            return Translate([random_height, random_width], 
                             lazy=True)(inputs[0])
        else:
            outputs = Translate([random_height, random_width],
                                 interp=self.interp)(*inputs)
            return outputs


class RandomChoiceTranslate(object):

    def __init__(self,
                 values,
                 p=None,
                 interp='bilinear',
                 lazy=False):
        """
        Randomly translate an image some fraction of total height and/or
        some fraction of total width from a list of potential values. 
        If the image has multiple channels,
        the same translation will be applied to each channel.

        Arguments
        ---------
        values : a list or tuple
            the values from which the translation value will be sampled

        p : a list or tuple the same length as `values`
            the probabilities of sampling any given value. Must sum to 1.

        interp : string in {'bilinear', 'nearest'} or list of strings
            type of interpolation to use. You can provide a different
            type of interpolation for each input, e.g. if you have two
            inputs then you can say `interp=['bilinear','nearest']

        lazy    : boolean
            if true, only create the affine transform matrix and return that
            if false, perform the transform on the tensor and return the tensor
        """
        if isinstance(values, (list, tuple)):
            values = th.FloatTensor(values)
        self.values = values
        if p is None:
            p = th.ones(len(values)) / len(values)
        else:
            if abs(1.0-sum(p)) > 1e-3:
                raise ValueError('Probs must sum to 1')
        self.p = p
        self.interp = interp
        self.lazy = lazy

    def __call__(self, *inputs):
        random_height = th_random_choice(self.values, p=self.p)
        random_width = th_random_choice(self.values, p=self.p)

        if self.lazy:
            return Translate([random_height, random_width],
                             lazy=True)(inputs[0])
        else:
            outputs = Translate([random_height, random_width],
                                interp=self.interp)(*inputs)
            return outputs


class Translate(object):

    def __init__(self, 
                 value, 
                 interp='bilinear',
                 lazy=False):
        """
        Arguments
        ---------
        value : float or 2-tuple of float
            if single value, both horizontal and vertical translation
            will be this value * total height/width. Thus, value should
            be a fraction of total height/width with range (-1, 1)

        interp : string in {'bilinear', 'nearest'} or list of strings
            type of interpolation to use. You can provide a different
            type of interpolation for each input, e.g. if you have two
            inputs then you can say `interp=['bilinear','nearest']

        """
        if not isinstance(value, (tuple,list)):
            value = (value, value)

        if value[0] > 1 or value[0] < -1:
            raise ValueError('Translation must be between -1 and 1')
        if value[1] > 1 or value[1] < -1:
            raise ValueError('Translation must be between -1 and 1')

        self.height_range = value[0]
        self.width_range = value[1]
        self.interp = interp
        self.lazy = lazy

    def __call__(self, *inputs):
        if not isinstance(self.interp, (tuple,list)):
            interp = [self.interp]*len(inputs)
        else:
            interp = self.interp

        tx = self.height_range * inputs[0].size(1)
        ty = self.width_range * inputs[0].size(2)

        translation_matrix = th.FloatTensor([[1, 0, tx],
                                             [0, 1, ty],
                                             [0, 0, 1]])
        if self.lazy:
            return translation_matrix
        else:
            outputs = []
            for idx, _input in enumerate(inputs):
                input_tf = th_affine2d(_input,
                                       translation_matrix,
                                       mode=interp[idx],
                                       center=True)
                outputs.append(input_tf)
            return outputs if idx > 1 else outputs[0]


class RandomShear(object):

    def __init__(self, 
                 shear_range,
                 interp='bilinear',
                 lazy=False):
        """
        Randomly shear an image with radians (-shear_range, shear_range)

        Arguments
        ---------
        shear_range : float
            radian bounds on the shear transform
        
        interp : string in {'bilinear', 'nearest'} or list of strings
            type of interpolation to use. You can provide a different
            type of interpolation for each input, e.g. if you have two
            inputs then you can say `interp=['bilinear','nearest']

        lazy    : boolean
            if false, perform the transform on the tensor and return the tensor
            if true, only create the affine transform matrix and return that
        """
        self.shear_range = shear_range
        self.interp = interp
        self.lazy = lazy

    def __call__(self, *inputs):
        shear = random.uniform(-self.shear_range, self.shear_range)
        if self.lazy:
            return Shear(shear, 
                         lazy=True)(inputs[0])
        else:
            outputs = Shear(shear,
                            interp=self.interp)(*inputs)
            return outputs


class RandomChoiceShear(object):

    def __init__(self,
                 values,
                 p=None,
                 interp='bilinear',
                 lazy=False):
        """
        Randomly shear an image with a value sampled from a list of values.

        Arguments
        ---------
        values : a list or tuple
            the values from which the rotation value will be sampled

        p : a list or tuple the same length as `values`
            the probabilities of sampling any given value. Must sum to 1.

        interp : string in {'bilinear', 'nearest'} or list of strings
            type of interpolation to use. You can provide a different
            type of interpolation for each input, e.g. if you have two
            inputs then you can say `interp=['bilinear','nearest']

        lazy    : boolean
            if false, perform the transform on the tensor and return the tensor
            if true, only create the affine transform matrix and return that
        """
        if isinstance(values, (list, tuple)):
            values = th.FloatTensor(values)
        self.values = values
        if p is None:
            p = th.ones(len(values)) / len(values)
        else:
            if abs(1.0-sum(p)) > 1e-3:
                raise ValueError('Probs must sum to 1')
        self.p = p
        self.interp = interp
        self.lazy = lazy

    def __call__(self, *inputs):
        shear = th_random_choice(self.values, p=self.p)

        if self.lazy:
            return Shear(shear, 
                         lazy=True)(inputs[0])
        else:
            outputs = Shear(shear,
                            interp=self.interp)(*inputs)
            return outputs 


class Shear(object):

    def __init__(self,
                 value,
                 interp='bilinear',
                 lazy=False):
        self.value = value
        self.interp = interp
        self.lazy = lazy

    def __call__(self, *inputs):
        if not isinstance(self.interp, (tuple,list)):
            interp = [self.interp]*len(inputs)
        else:
            interp = self.interp

        theta = (math.pi * self.value) / 180
        shear_matrix = th.FloatTensor([[1, -math.sin(theta), 0],
                                        [0, math.cos(theta), 0],
                                        [0, 0, 1]])
        if self.lazy:
            return shear_matrix
        else:
            outputs = []
            for idx, _input in enumerate(inputs):
                input_tf = th_affine2d(_input,
                                       shear_matrix,
                                       mode=interp[idx],
                                       center=True)
                outputs.append(input_tf)
            return outputs if idx > 1 else outputs[0]


class RandomZoom(object):

    def __init__(self, 
                 zoom_range,
                 interp='bilinear',
                 lazy=False):
        """
        Randomly zoom in and/or out on an image 

        Arguments
        ---------
        zoom_range : tuple or list with 2 values, both between (0, infinity)
            lower and upper bounds on percent zoom. 
            Anything less than 1.0 will zoom in on the image, 
            anything greater than 1.0 will zoom out on the image.
            e.g. (0.7, 1.0) will only zoom in, 
                 (1.0, 1.4) will only zoom out,
                 (0.7, 1.4) will randomly zoom in or out

        interp : string in {'bilinear', 'nearest'} or list of strings
            type of interpolation to use. You can provide a different
            type of interpolation for each input, e.g. if you have two
            inputs then you can say `interp=['bilinear','nearest']

        lazy    : boolean
            if false, perform the transform on the tensor and return the tensor
            if true, only create the affine transform matrix and return that
        """
        if not isinstance(zoom_range, list) and not isinstance(zoom_range, tuple):
            raise ValueError('zoom_range must be tuple or list with 2 values')
        self.zoom_range = zoom_range
        self.interp = interp
        self.lazy = lazy

    def __call__(self, *inputs):
        zx = random.uniform(self.zoom_range[0], self.zoom_range[1])
        zy = random.uniform(self.zoom_range[0], self.zoom_range[1])

        if self.lazy:
            return Zoom([zx, zy], lazy=True)(inputs[0])
        else:
            outputs = Zoom([zx, zy], 
                           interp=self.interp)(*inputs)
            return outputs


class RandomChoiceZoom(object):

    def __init__(self, 
                 values,
                 p=None,
                 interp='bilinear',
                 lazy=False):
        """
        Randomly zoom in and/or out on an image with a value sampled from
        a list of values

        Arguments
        ---------
        values : a list or tuple
            the values from which the applied zoom value will be sampled

        p : a list or tuple the same length as `values`
            the probabilities of sampling any given value. Must sum to 1.

        interp : string in {'bilinear', 'nearest'} or list of strings
            type of interpolation to use. You can provide a different
            type of interpolation for each input, e.g. if you have two
            inputs then you can say `interp=['bilinear','nearest']

        lazy    : boolean
            if false, perform the transform on the tensor and return the tensor
            if true, only create the affine transform matrix and return that
        """
        if isinstance(values, (list, tuple)):
            values = th.FloatTensor(values)
        self.values = values
        if p is None:
            p = th.ones(len(values)) / len(values)
        else:
            if abs(1.0-sum(p)) > 1e-3:
                raise ValueError('Probs must sum to 1')
        self.p = p
        self.interp = interp
        self.lazy = lazy

    def __call__(self, *inputs):
        zx = th_random_choice(self.values, p=self.p)
        zy = th_random_choice(self.values, p=self.p)

        if self.lazy:
            return Zoom([zx, zy], lazy=True)(inputs[0])
        else:
            outputs = Zoom([zx, zy], 
                           interp=self.interp)(*inputs)
            return outputs


class Zoom(object):

    def __init__(self,
                 value,
                 interp='bilinear',
                 lazy=False):
        """
        Arguments
        ---------
        value : float
            Fractional zoom.
            =1 : no zoom
            >1 : zoom-in (value-1)%
            <1 : zoom-out (1-value)%

        interp : string in {'bilinear', 'nearest'} or list of strings
            type of interpolation to use. You can provide a different
            type of interpolation for each input, e.g. if you have two
            inputs then you can say `interp=['bilinear','nearest']

        lazy: boolean
            If true, just return transformed
        """

        if not isinstance(value, (tuple,list)):
            value = (value, value)
        self.value = value
        self.interp = interp
        self.lazy = lazy

    def __call__(self, *inputs):
        if not isinstance(self.interp, (tuple,list)):
            interp = [self.interp]*len(inputs)
        else:
            interp = self.interp

        zx, zy = self.value
        zoom_matrix = th.FloatTensor([[zx, 0, 0],
                                      [0, zy, 0],
                                      [0, 0,  1]])        

        if self.lazy:
            return zoom_matrix
        else:
            outputs = []
            for idx, _input in enumerate(inputs):
                input_tf = th_affine2d(_input,
                                       zoom_matrix,
                                       mode=interp[idx],
                                       center=True)
                outputs.append(input_tf)
            return outputs if idx > 1 else outputs[0]






import datetime
import warnings

try:
    from inspect import signature
except:
    warnings.warn('inspect.signature not available... '
        'you should upgrade to Python 3.x')

import torch.nn.functional as F
import torch.optim as optim

from ..metrics import Metric, CategoricalAccuracy, BinaryAccuracy
from ..initializers import GeneralInitializer

def _add_regularizer_to_loss_fn(loss_fn, 
                                regularizer_container):
    def new_loss_fn(output_batch, target_batch):
        return loss_fn(output_batch, target_batch) + regularizer_container.get_value()
    return new_loss_fn

def _is_iterable(x):
    return isinstance(x, (tuple, list))
def _is_tuple_or_list(x):
    return isinstance(x, (tuple, list))

def _parse_num_inputs_and_targets_from_loader(loader):
    """ NOT IMPLEMENTED """
    #batch = next(iter(loader))
    num_inputs = loader.dataset.num_inputs
    num_targets = loader.dataset.num_targets
    return num_inputs, num_targets

def _parse_num_inputs_and_targets(inputs, targets=None):
    if isinstance(inputs, (list, tuple)):
        num_inputs = len(inputs)
    else:
        num_inputs = 1
    if targets is not None:
        if isinstance(targets, (list, tuple)):
            num_targets = len(targets)
        else:
            num_targets = 1
    else:
        num_targets = 0
    return num_inputs, num_targets

def _standardize_user_data(inputs, targets=None):
    if not isinstance(inputs, (list,tuple)):
        inputs = [inputs]
    if targets is not None:
        if not isinstance(targets, (list,tuple)):
            targets = [targets]
        return inputs, targets
    else:
        return inputs

def _validate_metric_input(metric):
    if isinstance(metric, str):
        if metric.upper() == 'CATEGORICAL_ACCURACY' or metric.upper() == 'ACCURACY':
            return CategoricalAccuracy()
        elif metric.upper() == 'BINARY_ACCURACY':
            return BinaryAccuracy()
        else:
            raise ValueError('Invalid metric string input - must match pytorch function.')
    elif isinstance(metric, Metric):
        return metric
    else:
        raise ValueError('Invalid metric input')

def _validate_loss_input(loss):
    dir_f = dir(F)
    loss_fns = [d.lower() for d in dir_f]
    if isinstance(loss, str):
        if loss.lower() == 'unconstrained':
            return lambda x: x
        elif loss.lower() == 'unconstrained_sum':
            return lambda x: x.sum()
        elif loss.lower() == 'unconstrained_mean':
            return lambda x: x.mean()
        else:
            try:
                str_idx = loss_fns.index(loss.lower())
            except:
                raise ValueError('Invalid loss string input - must match pytorch function.')
            return getattr(F, dir(F)[str_idx])
    elif callable(loss):
        return loss
    else:
        raise ValueError('Invalid loss input')

def _validate_optimizer_input(optimizer):
    dir_optim = dir(optim)
    opts = [o.lower() for o in dir_optim]
    if isinstance(optimizer, str):
        try:
            str_idx = opts.index(optimizer.lower())    
        except:
            raise ValueError('Invalid optimizer string input - must match pytorch function.')
        return getattr(optim, dir_optim[str_idx])
    elif hasattr(optimizer, 'step') and hasattr(optimizer, 'zero_grad'):
        return optimizer
    else:
        raise ValueError('Invalid optimizer input')

def _validate_initializer_input(initializer):
    if isinstance(initializer, str):
        try:
            initializer = GeneralInitializer(initializer)
        except:
            raise ValueError('Invalid initializer string input - must match pytorch function.')
        return initializer
    elif callable(initializer):
        return initializer
    else:
        raise ValueError('Invalid optimizer input')

def _get_current_time():
    return datetime.datetime.now().strftime("%B %d, %Y - %I:%M%p")

def _nb_function_args(fn):
    return len(signature(fn).parameters)


"""
Transforms to distort local or global information of an image
"""


import torch as th
import numpy as np
import random


class Scramble(object):
    """
    Create blocks of an image and scramble them
    """
    def __init__(self, blocksize):
        self.blocksize = blocksize

    def __call__(self, *inputs):
        outputs = []
        for idx, _input in enumerate(inputs):
            size = _input.size()
            img_height = size[1]
            img_width = size[2]

            x_blocks = int(img_height/self.blocksize) # number of x blocks
            y_blocks = int(img_width/self.blocksize)
            ind = th.randperm(x_blocks*y_blocks)

            new = th.zeros(_input.size())
            count = 0
            for i in range(x_blocks):
                for j in range (y_blocks):
                    row = int(ind[count] / x_blocks)
                    column = ind[count] % x_blocks
                    new[:, i*self.blocksize:(i+1)*self.blocksize, j*self.blocksize:(j+1)*self.blocksize] = \
                    _input[:, row*self.blocksize:(row+1)*self.blocksize, column*self.blocksize:(column+1)*self.blocksize]
                    count += 1
            outputs.append(new)
        return outputs if idx > 1 else outputs[0]
 

class RandomChoiceScramble(object):

    def __init__(self, blocksizes):
        self.blocksizes = blocksizes

    def __call__(self, *inputs):
        blocksize = random.choice(self.blocksizes)
        outputs = Scramble(blocksize=blocksize)(*inputs)
        return outputs


def _blur_image(image, H):
    # break image up into its color components
    size = image.shape
    imr = image[0,:,:]
    img = image[1,:,:]
    imb = image[2,:,:]

    # compute Fourier transform and frequqnecy spectrum
    Fim1r = np.fft.fftshift(np.fft.fft2(imr))
    Fim1g  = np.fft.fftshift(np.fft.fft2(img))
    Fim1b  = np.fft.fftshift(np.fft.fft2(imb))
    
    # Apply the lowpass filter to the Fourier spectrum of the image
    filtered_imager = np.multiply(H, Fim1r)
    filtered_imageg = np.multiply(H, Fim1g)
    filtered_imageb = np.multiply(H, Fim1b)
    
    newim = np.zeros(size)

    # convert the result to the spatial domain.
    newim[0,:,:] = np.absolute(np.real(np.fft.ifft2(filtered_imager)))
    newim[1,:,:] = np.absolute(np.real(np.fft.ifft2(filtered_imageg)))
    newim[2,:,:] = np.absolute(np.real(np.fft.ifft2(filtered_imageb)))

    return newim.astype('uint8')

def _butterworth_filter(rows, cols, thresh, order):
    # X and Y matrices with ranges normalised to +/- 0.5
    array1 = np.ones(rows)
    array2 = np.ones(cols)
    array3 = np.arange(1,rows+1)
    array4 = np.arange(1,cols+1)

    x = np.outer(array1, array4)
    y = np.outer(array3, array2)

    x = x - float(cols/2) - 1
    y = y - float(rows/2) - 1

    x = x / cols
    y = y / rows

    radius = np.sqrt(np.square(x) + np.square(y))

    matrix1 = radius/thresh
    matrix2 = np.power(matrix1, 2*order)
    f = np.reciprocal(1 + matrix2)

    return f


class Blur(object):
    """
    Blur an image with a Butterworth filter with a frequency
    cutoff matching local block size
    """
    def __init__(self, threshold, order=5):
        """
        scramble blocksize of 128 => filter threshold of 64
        scramble blocksize of 64 => filter threshold of 32
        scramble blocksize of 32 => filter threshold of 16
        scramble blocksize of 16 => filter threshold of 8
        scramble blocksize of 8 => filter threshold of 4
        """
        self.threshold = threshold
        self.order = order

    def __call__(self, *inputs):
        """
        inputs should have values between 0 and 255
        """
        outputs = []
        for idx, _input in enumerate(inputs):
            rows = _input.size(1)
            cols = _input.size(2)
            fc = self.threshold # threshold
            fs = 128.0 # max frequency
            n  = self.order # filter order
            fc_rad = (fc/fs)*0.5
            H = _butterworth_filter(rows, cols, fc_rad, n)
            _input_blurred = _blur_image(_input.numpy().astype('uint8'), H)
            _input_blurred = th.from_numpy(_input_blurred).float()
            outputs.append(_input_blurred)

        return outputs if idx > 1 else outputs[0]


class RandomChoiceBlur(object):

    def __init__(self, thresholds, order=5):
        """
        thresholds = [64.0, 32.0, 16.0, 8.0, 4.0]
        """
        self.thresholds = thresholds
        self.order = order

    def __call__(self, *inputs):
        threshold = random.choice(self.thresholds)
        outputs = Blur(threshold=threshold, order=self.order)(*inputs)
        return outputs










import torch as th
import math

class Sampler(object):
    """Base class for all Samplers.

    Every Sampler subclass has to provide an __iter__ method, providing a way
    to iterate over indices of dataset elements, and a __len__ method that
    returns the length of the returned iterators.
    """

    def __init__(self, data_source):
        pass

    def __iter__(self):
        raise NotImplementedError

    def __len__(self):
        raise NotImplementedError

class StratifiedSampler(Sampler):
    """Stratified Sampling

    Provides equal representation of target classes in each batch
    """
    def __init__(self, class_vector, batch_size):
        """
        Arguments
        ---------
        class_vector : torch tensor
            a vector of class labels
        batch_size : integer
            batch_size
        """
        self.n_splits = int(class_vector.size(0) / batch_size)
        self.class_vector = class_vector

    def gen_sample_array(self):
        try:
            from sklearn.model_selection import StratifiedShuffleSplit
        except:
            print('Need scikit-learn for this functionality')
        import numpy as np
        
        s = StratifiedShuffleSplit(n_splits=self.n_splits, test_size=0.5)
        X = th.randn(self.class_vector.size(0),2).numpy()
        y = self.class_vector.numpy()
        s.get_n_splits(X, y)

        train_index, test_index = next(s.split(X, y))
        return np.hstack([train_index, test_index])

    def __iter__(self):
        return iter(self.gen_sample_array())

    def __len__(self):
        return len(self.class_vector)

class MultiSampler(Sampler):
    """Samples elements more than once in a single pass through the data.

    This allows the number of samples per epoch to be larger than the number
    of samples itself, which can be useful when training on 2D slices taken
    from 3D images, for instance.
    """
    def __init__(self, nb_samples, desired_samples, shuffle=False):
        """Initialize MultiSampler

        Arguments
        ---------
        data_source : the dataset to sample from
        
        desired_samples : number of samples per batch you want
            whatever the difference is between an even division will
            be randomly selected from the samples.
            e.g. if len(data_source) = 3 and desired_samples = 4, then
            all 3 samples will be included and the last sample will be
            randomly chosen from the 3 original samples.

        shuffle : boolean
            whether to shuffle the indices or not
        
        Example:
            >>> m = MultiSampler(2, 6)
            >>> x = m.gen_sample_array()
            >>> print(x) # [0,1,0,1,0,1]
        """
        self.data_samples = nb_samples
        self.desired_samples = desired_samples
        self.shuffle = shuffle

    def gen_sample_array(self):
        from trainingtorchsample.utils import th_random_choice
        n_repeats = self.desired_samples / self.data_samples
        cat_list = []
        for i in range(math.floor(n_repeats)):
            cat_list.append(th.arange(0,self.data_samples))
        # add the left over samples
        left_over = self.desired_samples % self.data_samples
        if left_over > 0:
            cat_list.append(th_random_choice(self.data_samples, left_over))
        self.sample_idx_array = th.cat(cat_list).long()
        return self.sample_idx_array

    def __iter__(self):
        return iter(self.gen_sample_array())

    def __len__(self):
        return self.desired_samples


class SequentialSampler(Sampler):
    """Samples elements sequentially, always in the same order.

    Arguments:
        data_source (Dataset): dataset to sample from
    """

    def __init__(self, nb_samples):
        self.num_samples = nb_samples

    def __iter__(self):
        return iter(range(self.num_samples))

    def __len__(self):
        return self.num_samples


class RandomSampler(Sampler):
    """Samples elements randomly, without replacement.

    Arguments:
        data_source (Dataset): dataset to sample from
    """

    def __init__(self, nb_samples):
        self.num_samples = nb_samples

    def __iter__(self):
        return iter(th.randperm(self.num_samples).long())

    def __len__(self):
        return self.num_samples





"""
Classes to initialize module weights
"""

from fnmatch import fnmatch

import torch.nn.init


def _validate_initializer_string(init):
    dir_f = dir(torch.nn.init)
    loss_fns = [d.lower() for d in dir_f]
    if isinstance(init, str):
        try:
            str_idx = loss_fns.index(init.lower())
        except:
            raise ValueError('Invalid loss string input - must match pytorch function.')
        return getattr(torch.nn.init, dir(torch.nn.init)[str_idx])
    elif callable(init):
        return init
    else:
        raise ValueError('Invalid loss input')


class InitializerContainer(object):

    def __init__(self, initializers):
        self._initializers = initializers

    def apply(self, model):
        for initializer in self._initializers:
            model.apply(initializer)


class Initializer(object):

    def __call__(self, module):
        raise NotImplementedError('Initializer must implement this method')


class GeneralInitializer(Initializer):

    def __init__(self, initializer, bias=False, bias_only=False, **kwargs):
        self._initializer = _validate_initializer_string(initializer)
        self.kwargs = kwargs

    def __call__(self, module):
        classname = module.__class__.__name__
        if fnmatch(classname, self.module_filter) and hasattr(module, 'weight'):
            if self.bias_only:
                self._initializer(module.bias.data, **self.kwargs)
            else:
                self._initializer(module.weight.data, **self.kwargs)
                if self.bias:
                    self._initializer(module.bias.data, **self.kwargs)


class Normal(Initializer):

    def __init__(self, mean=0.0, std=0.02, bias=False, 
                 bias_only=False, module_filter='*'):
        self.mean = mean
        self.std = std

        self.bias = bias
        self.bias_only = bias_only
        self.module_filter = module_filter

        super(Normal, self).__init__()

    def __call__(self, module):
        classname = module.__class__.__name__
        if fnmatch(classname, self.module_filter) and hasattr(module, 'weight'):
            if self.bias_only:
                torch.nn.init.normal(module.bias.data, mean=self.mean, std=self.std)
            else:
                torch.nn.init.normal(module.weight.data, mean=self.mean, std=self.std)
                if self.bias:
                    torch.nn.init.normal(module.bias.data, mean=self.mean, std=self.std)


class Uniform(Initializer):

    def __init__(self, a=0, b=1, bias=False, bias_only=False, module_filter='*'):
        self.a = a
        self.b = b

        self.bias = bias
        self.bias_only = bias_only
        self.module_filter = module_filter

        super(Uniform, self).__init__()

    def __call__(self, module):
        classname = module.__class__.__name__
        if fnmatch(classname, self.module_filter) and hasattr(module, 'weight'):
            if self.bias_only:
                torch.nn.init.uniform(module.bias.data, a=self.a, b=self.b)
            else:
                torch.nn.init.uniform(module.weight.data, a=self.a, b=self.b)
                if self.bias:
                    torch.nn.init.uniform(module.bias.data, a=self.a, b=self.b)


class ConstantInitializer(Initializer):

    def __init__(self, value, bias=False, bias_only=False, module_filter='*'):
        self.value = value

        self.bias = bias
        self.bias_only = bias_only
        self.module_filter = module_filter

        super(ConstantInitializer, self).__init__()

    def __call__(self, module, bias=False, bias_only=False, module_filter='*'):
        classname = module.__class__.__name__
        if fnmatch(classname, self.module_filter) and hasattr(module, 'weight'):
            if self.bias_only:
                torch.nn.init.constant(module.bias.data, val=self.value)
            else:
                torch.nn.init.constant(module.weight.data, val=self.value)
                if self.bias:
                    torch.nn.init.constant(module.bias.data, val=self.value)


class XavierUniform(Initializer):

    def __init__(self, gain=1, bias=False, bias_only=False, module_filter='*'):
        self.gain = gain

        self.bias = bias
        self.bias_only = bias_only
        self.module_filter = module_filter

        super(XavierUniform, self).__init__()

    def __call__(self, module):
        classname = module.__class__.__name__
        if fnmatch(classname, self.module_filter) and hasattr(module, 'weight'):
            if self.bias_only:
                torch.nn.init.xavier_uniform(module.bias.data, gain=self.gain)
            else:
                torch.nn.init.xavier_uniform(module.weight.data, gain=self.gain)
                if self.bias:
                    torch.nn.init.xavier_uniform(module.bias.data, gain=self.gain)


class XavierNormal(Initializer):

    def __init__(self, gain=1, bias=False, bias_only=False, module_filter='*'):
        self.gain = gain
        
        self.bias = bias
        self.bias_only = bias_only
        self.module_filter = module_filter

        super(XavierNormal, self).__init__()

    def __call__(self, module):
        classname = module.__class__.__name__
        if fnmatch(classname, self.module_filter) and hasattr(module, 'weight'):
            if self.bias_only:
                torch.nn.init.xavier_normal(module.bias.data, gain=self.gain)
            else:
                torch.nn.init.xavier_normal(module.weight.data, gain=self.gain)
                if self.bias:
                    torch.nn.init.xavier_normal(module.bias.data, gain=self.gain)


class KaimingUniform(Initializer):

    def __init__(self, a=0, mode='fan_in', bias=False, bias_only=False, module_filter='*'):
        self.a = a
        self.mode = mode
        
        self.bias = bias
        self.bias_only = bias_only
        self.module_filter = module_filter

        super(KaimingUniform, self).__init__()

    def __call__(self, module):
        classname = module.__class__.__name__
        if fnmatch(classname, self.module_filter) and hasattr(module, 'weight'):
            if self.bias_only:
                torch.nn.init.kaiming_uniform(module.bias.data, a=self.a, mode=self.mode)
            else:
                torch.nn.init.kaiming_uniform(module.weight.data, a=self.a, mode=self.mode)
                if self.bias:
                    torch.nn.init.kaiming_uniform(module.bias.data, a=self.a, mode=self.mode)


class KaimingNormal(Initializer):

    def __init__(self, a=0, mode='fan_in', bias=False, bias_only=False, module_filter='*'):
        self.a = a
        self.mode = mode
        
        self.bias = bias
        self.bias_only = bias_only
        self.module_filter = module_filter

        super(KaimingNormal, self).__init__()

    def __call__(self, module):
        classname = module.__class__.__name__
        if fnmatch(classname, self.module_filter) and hasattr(module, 'weight'):
            if self.bias_only:
                torch.nn.init.kaiming_normal(module.bias.data, a=self.a, mode=self.mode)
            else:
                torch.nn.init.kaiming_normal(module.weight.data, a=self.a, mode=self.mode)
                if self.bias:
                    torch.nn.init.kaiming_normal(module.bias.data, a=self.a, mode=self.mode)


class Orthogonal(Initializer):

    def __init__(self, gain=1, bias=False, bias_only=False, module_filter='*'):
        self.gain = gain
        
        self.bias = bias
        self.bias_only = bias_only
        self.module_filter = module_filter

        super(Orthogonal, self).__init__()

    def __call__(self, module):
        classname = module.__class__.__name__
        if fnmatch(classname, self.module_filter) and hasattr(module, 'weight'):
            if self.bias_only:
                torch.nn.init.orthogonal(module.bias.data, gain=self.gain)
            else:
                torch.nn.init.orthogonal(module.weight.data, gain=self.gain)
                if self.bias:
                    torch.nn.init.orthogonal(module.bias.data, gain=self.gain)


class Sparse(Initializer):

    def __init__(self, sparsity, std=0.01, bias=False, bias_only=False, module_filter='*'):
        self.sparsity = sparsity
        self.std = std
        
        self.bias = bias
        self.bias_only = bias_only
        self.module_filter = module_filter

        super(Sparse, self).__init__()

    def __call__(self, module):
        classname = module.__class__.__name__
        if fnmatch(classname, self.module_filter) and hasattr(module, 'weight'):
            if self.bias_only:
                torch.nn.init.sparse(module.bias.data, sparsity=self.sparsity, std=self.std)
            else:
                torch.nn.init.sparse(module.weight.data, sparsity=self.sparsity, std=self.std)
                if self.bias:
                    torch.nn.init.sparse(module.bias.data, sparsity=self.sparsity, std=self.std)



